#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
2020, 2023 - Rafael Guerreiro Osorio - Ipea.

pynad - métodos de identificação

"""
import pyarrow as pa
import pyarrow.compute as pc
from .auxiliares import elapsed, CRCLR, UPCLR


IDVARS = ['ano', 'trimestre', 'upa', 'v1008', 'v1016',
          'v2001', 'v2003', 'v2005', 'v2007', 'v2008',
          'v20081', 'v20082', 'v2009']
KEYS = ['upa', 'v1008', 'v2003', 'v1016']


def cli_parcial(count, total, pid, cli_args):
    """Atualiza CLI."""
    print(CRCLR, UPCLR * 4)
    print(f'Montando paineis - {cli_args[0]} de {cli_args[1]} -',
          f'{elapsed(cli_args[2])}\n')
    print(f'PAINEL {pid}:')
    print(f'Identificando o painel - {count * 100 / total:.1f}%',
          f'(UPA {count} de {total})', end='', flush=True)


def identifica_painel(pid, table, cli_args):
    """Identifica um painel recebido como pyarrow table."""
    tabpid = table.select(IDVARS)
    classificadores = (__id_cls1__, __id_cls2__, __id_cls3__)
    upas = tuple(value[0].as_py() for value in
                 pc.value_counts(tabpid['upa']))
    uparows = []
    for idx, upa in enumerate(upas):
        cli_parcial(idx + 1, len(upas), pid, cli_args)
        doms = tuple(value[0].as_py() for value in
                     pc.value_counts(tabpid.filter(
                         pc.field('upa') == upa)['v1008']))
        for dom in doms:
            pessoas = tabpid.filter((pc.field('upa') == upa) &
                                    (pc.field('v1008') == dom)).to_pylist()
            pessoas = load_persons(pid, pessoas)
            # alguma pessoa tem data de nascimento ignorada
            if any(pessoa.dtd == 99 for pessoa in pessoas):
                __id_birthdate_unknown__(pessoas)

            # identificar e separar os grupos domésticos do domicílio
            grpsdom = __id_split_grpds__(pessoas)

            # identificar os indivíduos de cada grupo doméstico
            for grpdom in grpsdom:
                for classificador in classificadores:
                    classificadas = classificador(grpsdom[grpdom],
                                                  grpdom)
                    if classificadas:
                        break
                for pessoa in classificadas:
                    uparows.append(pessoa.classificada())
    dtypes = [pa.field('upa', pa.int32()),
              pa.field('pid', pa.int32()),
              pa.field('piddna', pa.int32())]
    dtypes.extend([pa.field(var, pa.int8()) for var in
                   ('v1008', 'v1016', 'v2003', 'pidcla',
                    'pidgrp', 'pidgrpent', 'pidind', 'pidindent',
                    'piddnd', 'piddnm')])
    dtypes = pa.schema(dtypes)
    return table.join(pa.Table.from_pylist(uparows, schema=dtypes), KEYS)


def __id_aggr_persons__(pessoas, estrita):
    """
    Agrega pessoas em indivíduos.

    **pessoas** são os registros das diferentes pessoas de todas
    as entrevistas realizadas em um domicílio

    **estrita** = True considera sexo, data de nascimento
    número de ordem e situação no domicílio

    **estrita** = False considera sexo e data de nascimento

    Retorna um dicionário de indivíduos com duas listas, a das entrevistas
    do indivíduo, e a das pessoas que lhe correspondem em cada entrevista
    """
    individuos = {}
    for pessoa in pessoas:
        identidade = pessoa.ident(estrita)
        if identidade not in individuos:
            individuos[identidade] = [[], []]
        individuos[identidade][0].append(pessoa.ent)
        individuos[identidade][1].append(pessoa)

    # se apenas sexo e data de nascimento, precisa controlar gemeos,
    # que aparecem como um indivíduo com entrevistas duplicadas
    # é preciso separá-los
    if not estrita:
        for individuo in list(individuos.keys()):
            reflist = list(set(individuos[individuo][0]))
            reflist.sort()
            if individuos[individuo][0] != reflist:
                for entrevista in reflist:
                    gemeos = [idx for idx, item in
                              enumerate(individuos[individuo][0])
                              if item == entrevista]
                    for gemeo, ordem in enumerate(gemeos):
                        identidade = individuo + 'gemeo' + str(gemeo + 1)
                        if identidade not in individuos:
                            individuos[identidade] = [[], []]
                        individuos[identidade][0].append(
                            individuos[individuo][1][ordem].ent)
                        individuos[identidade][1].append(
                            individuos[individuo][1][ordem])
                del individuos[individuo]

    # lista para conjunto de entrevistas do indivíduo
    for individuo in individuos:
        individuos[individuo][0] = list(set(individuos[individuo][0]))
    return individuos


def __id_birthdate_unknown__(pessoas):
    """
    Trata datas de nascimento ignoradas.

    Verifica se uma pessoa com data de nascimento ignorada em uma
    entrevista teve data declarada em outra, imputando a data em caso
    afirmativo
    """
    # separa as pessoas segundo a situação da data de nascimento
    dataign = [pessoa for pessoa in pessoas if pessoa.dtd == 99]
    datacon = [pessoa for pessoa in pessoas if pessoa.dtd != 99]
    for ign in dataign:

        # para cada pessoa com data ignorada, a lista de doadores
        # inicial considera todas as com data conhecida em outras
        # entrevistas, do mesmo sexo e, e com diferença de até
        # 3 anos do ano de nascimento estimado para a data ignorada
        doadores = [pessoa for pessoa in datacon
                    if pessoa.ent != ign.ent
                    and pessoa.sex == ign.sex
                    and abs(pessoa.dta - ign.dta) <= 3]

        # preferência para doadores com a mesma situação no domicílio
        # se existirem, desprezar os demais
        if [pessoa for pessoa in doadores if pessoa.std == ign.std]:
            doadores = [pessoa for pessoa in doadores
                        if pessoa.std == ign.std]

        # se não há doadores com a mesma situação no domicílio,
        # verificar se há doadores do conjunto permitido de troca
        # de código de situação, eliminando os que não estejam
        elif doadores:
            filtro = []
            if ign.std in (1, 2, 3):
                filtro = (1, 2, 3)
            elif ign.std in (4, 5, 6):
                filtro = (4, 5, 6)
            elif ign.std in (8, 9):
                filtro = (8, 9)
            else:
                filtro = (ign.std, )
            doadores = [doador for doador in doadores
                        if doador.std in filtro]

        # havendo doadores, a lista é ordenada pela proximidade
        # do ano de nascimento, sendo escolhido o mais próximo
        # o primeiro da lista
        if doadores:
            difs = [abs(doador.dta - ign.dta) for doador in doadores]
            mindif = [idx for idx, dif in enumerate(difs)
                      if dif == min(difs)][0]
            doador = doadores[mindif]
            ign.igd = doador.dtd
            ign.igm = doador.dtm
            ign.iga = doador.dta

    # imputa as datas de nascimento para as pessoas com data ignorada que
    # encontraram doadoras
    for pessoa in pessoas:
        if pessoa.igd:
            pessoa.dtd = pessoa.igd
            pessoa.dtm = pessoa.igm
            pessoa.dta = pessoa.iga
    return pessoas


def __id_cls1__(pessoas, grpdom):
    """Classificador 1, grupo doméstico sem mudanças."""
    # verifica se o tamanho do grupo doméstico é o mesmo
    # em todas as entrevistas e gera sua lista de entrevistas
    tamanho = pessoas[0].tam
    entdom = []
    for pessoa in pessoas:
        if pessoa.tam != tamanho:
            return []
        if pessoa.ent not in entdom:
            entdom.append(pessoa.ent)
    entdom.sort()

    # agrega as pessoas em indivíduos
    # estrita=True: sexo, data, ordem e sitdom
    individuos = __id_aggr_persons__(pessoas, True)

    # verifica se indivíduos aparecem em todas as entrevistas do grupo
    for individuo in individuos:
        if individuos[individuo][0] != entdom:
            return []

    # identifica as pessoas
    for idx, individuo in enumerate(individuos):
        for pessoa in pessoas:
            identidade = pessoa.ident(True)
            if identidade == individuo:
                pessoa.cla = 1
                pessoa.grp = grpdom
                pessoa.ind = idx + 1
                pessoa.indent = sum(2 ** (item - 1) for item in
                                    individuos[individuo][0])
                pessoa.grpent = sum(2 ** (item - 1) for item in
                                    entdom)
    return pessoas


def __id_cls2__(pessoas, grpdom):
    """Classificador 2 grupos domésticos com tamanho constante e mudanças."""
    # verifica se o tamanho do grupo doméstico é o mesmo
    # em todas as entrevistas e gera sua lista de entrevistas
    tamanho = pessoas[0].tam
    entdom = []
    for pessoa in pessoas:
        if pessoa.tam != tamanho:
            return []
        if pessoa.ent not in entdom:
            entdom.append(pessoa.ent)
    entdom.sort()

    # agrega as pessoas em indivíduos
    # estrita=False: sexo e data
    individuos = __id_aggr_persons__(pessoas, False)

    # verifica se indivíduos aparecem em todas as entrevistas do grupo
    for individuo in individuos:
        if individuos[individuo][0] != entdom:
            return []

    # identifica as pessoas controlando gemeos do mesmo sexo
    indents = {}
    for individuo in individuos:
        indents[individuo] = []
    for idx, individuo in enumerate(individuos):
        for pessoa in pessoas:
            if (pessoa in individuos[individuo][1]
                    and pessoa.ent not in indents[individuo]):
                pessoa.cla = 2
                pessoa.grp = grpdom
                pessoa.ind = idx + 1
                pessoa.indent = sum(2 ** (item - 1) for item in
                                    individuos[individuo][0])
                pessoa.grpent = sum(2 ** (item - 1) for item in
                                    entdom)
                indents[individuo].append(pessoa.ent)
    return pessoas


def __id_cls3__(pessoas, grpdom):
    """Classificador 3 grupos domésticos com mudanças variadas."""
    # recursiva para gerar todas as permutações de individuos
    # que não aparecem na mesma entrevista
    def __permuta__(indids, entrevistas, outras, possiveis):
        for outra in outras:
            entconjunta = entrevistas + individuos[outra][0]
            entconjunta.sort()

            possiveis.append([entconjunta, indids[:] + [outra]])
            if entconjunta != entdom:
                mais = [out for out in outras
                        if not any(ent in entconjunta
                                   for ent in individuos[out][0])]
                if mais:
                    possiveis = __permuta__(indids[:] + [outra],
                                            entconjunta[:], mais,
                                            possiveis)
        return possiveis

    # agrega indívíduos
    def __aggr_ind__(juntar, cla):
        individuos[juntar[1][0]][0] = juntar[0][:]
        for outra in juntar[1][1:]:
            individuos[juntar[1][0]][1].extend(
                individuos[outra][1])
            del individuos[outra]
        for pessoa in individuos[juntar[1][0]][1]:
            pessoa.cla = cla
        return juntar[1]

    # entrevistas do grupo doméstico
    entdom = []
    for pessoa in pessoas:
        if pessoa.ent not in entdom:
            entdom.append(pessoa.ent)
    entdom.sort()

    # agrega as pessoas em indivíduos
    # estrita=False: sexo e data
    individuos = __id_aggr_persons__(pessoas, False)

    # separa e classifica os indivíduos que aparecem em todas as entrevistas
    todas = [individuo for individuo in individuos
             if individuos[individuo][0] == entdom]
    for individuo in todas:
        for pessoa in individuos[individuo][1]:
            pessoa.cla = 3

    # separa e classifica provisoriamente os que aparecem em algumas
    algumas = [individuo for individuo in individuos
               if individuo not in todas]
    for individuo in algumas:
        for pessoa in individuos[individuo][1]:
            pessoa.cla = 4

    # permutações de indivíduos sem aparições conjuntas
    possibilidades = []
    for individuo in algumas:
        entrevistas = list(individuos[individuo][0])
        outras = [outra for outra in list(algumas)
                  if outra != individuo
                  and not any(entrevista in entrevistas
                              for entrevista in individuos[outra][0])]
        if outras:
            possibilidades.extend(__permuta__([individuo],
                                              entrevistas,
                                              outras, []))

    # reduzir permutações a combinações
    organizadas = []
    for possibilidade in possibilidades:
        if not organizadas:
            organizadas.append(possibilidade)
        elif possibilidade not in organizadas:
            adicionar = True
            for organizada in organizadas:
                if (organizada[0] == possibilidade[0]
                        and len(organizada[1]) == len(possibilidade[1])
                        and all(individuo in organizada[1]
                                for individuo in possibilidade[1])):
                    adicionar = False
                    break
            if adicionar:
                organizadas.append(possibilidade)

    # comparar todos os pares de atributos e registrar diferenças
    # a ordem de atribs é relevante
    atribs = ('sex', 'dtd', 'dtm', 'dta', 'std')
    for idx, organizada in enumerate(organizadas):
        organizadas[idx].append([])
        organizadas[idx].append([])
        organizadas[idx].append([])
        possiveis = organizada[1][:]
        while possiveis:
            possivel = possiveis.pop(0)
            pessoa1 = individuos[possivel][1][0]
            outras = [outra for outra in possiveis
                      if outra != possivel]
            for outra in outras:
                pessoa2 = individuos[outra][1][0]
                difs = [atrib for atrib in atribs
                        if pessoa1.__dict__[atrib] !=
                        pessoa2.__dict__[atrib]]
                organizadas[idx][2].append(difs)
                organizadas[idx][3].append([pessoa1.std, pessoa2.std])
                organizadas[idx][4].append([pessoa1.dta, pessoa2.dta])

    # ordenar as combinações para processar primeiro as
    # que resultariam em fusão de mais fragmentos e maior número
    # de entrevistas
    organizadas.sort(key=lambda item: len(item[0]) * 100 + len(item[1]) * 10
                     + 5 - max([len(difs) for difs in item[2]]),
                     reverse=True)

    # conjunto de erros permitidos
    erros = [['sex'],
             ['dtd'],
             ['dtm'],
             ['dta'],
             ['sex', 'dtd'],
             ['sex', 'dtm'],
             ['sex', 'dta'],
             ['dtd', 'dtm'],
             ['dtm', 'dta'],
             ['dtd', 'dta']]

    # com mudança de condição no domicílio
    erstd = [erro[:] + ['std'] for erro in erros]

    # agregar fragmentos de indivíduos
    juntados = []
    residuos = []
    for organizada in organizadas:
        if not any(individuo in organizada[1] for individuo in juntados):

            # fragmentos com mesma condição no domicílio
            # diferença de até 3 anos entre anos de nascimento
            if (all(erro in erros for erro in organizada[2])
                    and all(abs(dta[0] - dta[1]) <= 3
                            for dta in organizada[4])):
                juntados.extend(__aggr_ind__(organizada, 6))

            # fragmentos com diferente condição no domicílio
            # diferença de até 3 anos entre anos de nascimento
            elif (all(erro in erstd for erro in organizada[2])
                  and all(abs(dta[0] - dta[1]) <= 3
                          for dta in organizada[4])):
                juntados.extend(__aggr_ind__(organizada, 7))
            else:
                residuos.append(organizada)
                for individuo in organizada[1]:
                    for pessoa in individuos[individuo][1]:
                        pessoa.cla = 5

    # identifica as pessoas controlando gemeos do mesmo sexo
    indents = {}
    for individuo in individuos:
        indents[individuo] = []
    for idx, individuo in enumerate(individuos):
        for pessoa in pessoas:
            if (pessoa in individuos[individuo][1]
                    and pessoa.ent not in indents[individuo]):
                pessoa.grp = grpdom
                pessoa.ind = idx + 1
                pessoa.indent = sum(2 ** (item - 1) for item in
                                    individuos[individuo][0])
                pessoa.grpent = sum(2 ** (item - 1) for item in
                                    entdom)
                indents[individuo].append(pessoa.ent)
    return pessoas


def __id_split_grpds__(pessoas):
    """Identifica e separa os grupos domésticos de um domicílio."""
    # dicionário de grupos domésticos
    grpsdom = {}

    # separa as pessoas por entrevista
    ents = {}
    for pessoa in pessoas:
        if pessoa.ent not in ents:
            ents[pessoa.ent] = []
        ents[pessoa.ent].append(pessoa)

    # uma entrevista, um grupo; nada a separar...
    if len(ents) == 1:
        grpsdom[1] = pessoas
        for pessoa in grpsdom[1]:
            pessoa.grp = 1
        return grpsdom

    # se mais de uma entrevista
    # agrega as pessoas em indivíduos
    # estrita=False: sexo e, data
    individuos = __id_aggr_persons__(pessoas, False)

    # procurar a entrevista com mais pessoas com data de nascimento conhecida
    inicio = [(ent, len([pessoa for pessoa in ents[ent]
                         if pessoa.dtd < 99]))
              for ent in ents]
    inicio.sort(key=lambda e: e[1], reverse=True)
    ent = inicio[0][0]
    curdom = 0

    # processar entrevistas e ir retirando as atribuídas a um grupo
    # doméstico da lista de entrevistas
    while ents:

        # atribui ao primeiro grupo as pessoas da entrevista inicial
        # e a retira da lista
        curdom += 1
        grpsdom[curdom] = ents[ent]
        del ents[ent]

        # achar outras entrevistas com pessoas de indivíduos que estavam
        # na entrevista inicial e removê-las
        for individuo in individuos.values():
            if any(pessoa in grpsdom[curdom] for pessoa in individuo[1]):
                for ent in individuo[0]:
                    if ent in ents:
                        grpsdom[curdom].extend(ents[ent])
                        del ents[ent]
                    for ini in inicio:
                        if ini[0] == ent:
                            inicio.remove(ini)
                    if not ents:
                        break
            if not ents:
                break

        # verifica se há entrevistas não atribuídas, havendo
        # seleciona a com maior número de pessoas e reinicia o loop
        # criando outro grupo doméstico
        if inicio:
            ent = inicio[0][0]

    # reordenar pessoas
    reord = []
    for grpdom in grpsdom:
        grpsdom[grpdom].sort(key=lambda pessoa: pessoa.ent * 100 + pessoa.ord)
        reord.append([min([pessoa.ent for pessoa in grpsdom[grpdom]]), grpdom])
    reord.sort(key=lambda e: e[0])

    # ordenar grupos domésticos segundo a visita
    # o grupo 1 é o da primeira entrevista
    grpsdomord = {}
    for idx, val in enumerate(reord):
        grpsdomord[idx + 1] = grpsdom[val[1]]
    return grpsdomord


class Person():
    """
    representa um registro de pessoa.

    As instâncias desta classe representam registros de pessoas e
    seus atributos correspondem às variáveis empregadas na identificação,
    especificadas em __load_panel_metadata__(), função que retorna
    **attribs** e **categs** usados para criar uma instância;
    **values** é o registro como lista de valores e os
    códigos dos atributos da pessoa
    """

    def __init__(self, attribs, categs):
        for key, value in attribs.items():
            self.__dict__[key] = value
        self.sexcat = categs['sexcat'][str(self.sex)]
        self.stdcat = categs['stdcat'][str(self.std)] + 18 * ' '
        self.pid = 0
        self.cla = 0
        self.grp = 0
        self.ind = 0
        self.igi = 0
        self.igd = 0
        self.igm = 0
        self.iga = 0
        self.grpent = 0
        self.indent = 0

    def __repr__(self):
        """
        Retorna uma string que representa a pessoa.

        Contém os atributos usados na identificação,
        bem como os resultantes
        """
        repres = (f'e{self.ent} {self.sexcat[:3]} '
                  + f'{self.dta}-{self.dtm:02.0f}-{self.dtd:02.0f} '
                  + f'{self.ord:02.0f} {self.stdcat[:18]} '
                  + f'cla {self.cla} grpd {self.grp} ind {self.ind:02.0f}')
        if self.igd:
            repres += f' ({self.age} anos - {self.ano - self.age}-99-99)'
        return repres

    def classificada(self):
        """
        Retorna o registro da pessoa.

        como dicionário python para a tabela de identificação
        """
        names = ('upa', 'v1008', 'v1016', 'v2003', 'pid', 'pidcla',
                 'pidgrp', 'pidgrpent', 'pidind', 'pidindent',
                 'piddnd', 'piddnm', 'piddna')
        keys = ('upa', 'dom', 'ent', 'ord', 'pid', 'cla',
                'grp', 'grpent', 'ind', 'indent', 'igd', 'igm', 'iga')
        return {names[idx]: self.__dict__[keys[idx]]
                for idx in range(len(names))}

    def ident(self, estrita):
        """Retorna os atributos para a agregação em indivíduos."""
        idnt = f'{self.sexcat[:3]} '
        idnt += f'{self.dta}-{self.dtm:02.0f}-{self.dtd:02.0f}'
        if estrita:
            idnt += f' {self.ord:02.0f} {self.stdcat[:18]}'
        return idnt


def load_persons(pid, pessoas):
    """Carrega uma pessoa."""
    v2005 = {'1': 'pessoa responsável pelo domicílio',
             '2': 'cônjuge ou companheiro(a) de sexo diferente',
             '3': 'cônjuge ou companheiro(a) do mesmo sexo',
             '4': 'filho(a) do responsável e do cônjuge',
             '5': 'filho(a) somente do responsável',
             '6': 'enteado(a)',
             '7': 'genro ou nora',
             '8': 'pai, mãe, padrasto ou madrasta',
             '9': 'sogro(a)',
             '10': 'neto(a)',
             '11': 'bisneto(a)',
             '12': 'irmão ou irmã',
             '13': 'avô ou avó',
             '14': 'outro parente',
             '15': 'agregado(a) - não parente que não compartilha despesas',
             '16': 'convivente - não parente que compartilha despesas',
             '17': 'pensionista',
             '18': 'empregado(a) doméstico(a)',
             '19': 'parente do(a) empregado(a) doméstico(a)'}
    persons = []
    for pessoa in pessoas:
        attribs = {'ano': pessoa['ano'],
                   'tri': pessoa['trimestre'],
                   'upa': pessoa['upa'],
                   'dom': pessoa['v1008'],
                   'ent': pessoa['v1016'],
                   'tam': pessoa['v2001'],
                   'ord': pessoa['v2003'],
                   'std': pessoa['v2005'],
                   'sex': pessoa['v2007'],
                   'dtd': pessoa['v2008'],
                   'dtm': pessoa['v20081'],
                   'dta': pessoa['v20082'],
                   'age': pessoa['v2009']}
        categs = {'stdcat': v2005,
                  'sexcat': {'1': 'homem', '2': 'mulher'}}
        persons.append(Person(attribs, categs))
        persons[-1].pid = pid
        if persons[-1].dtd == 99:
            persons[-1].dta = int(persons[-1].pid / 10) - persons[-1].age
    return persons



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Pynad Command Line Interface.

pynad v3
Ipea - 2023
Rafael Guerreiro Osorio

    def _menu_help(self):
        import webbrowser
        webbrowser.open(str(Path(APP_PATH, 'help.html')))
"""
from json import dumps, loads
from pathlib import Path
from platform import system
from shutil import rmtree
from time import sleep
import tkinter as tk
from tkinter import filedialog
from .auxiliares import elapsed, continuar, pausar, cli_main
from .auxiliares import TCOLS, ARROW, HEADER_F, HEADER_B, HEADER_W, COPIA_F
from .auxiliares import COPIA_B, COPIA_H, COPIA_W, UPCLR
from .copia_local import register, verify, sync
from .copia_local import COPIA_LOCAL, IDFILE, LSFILE, MPFILE
from .converter import sync_microdados, MICRO
from .metadados import generate, META
from .paineis import sync_paineis, PAINEIS


# ativar terminal escape sequences no Windows
if system() == 'Windows':
    from ctypes import windll
    windll.kernel32.SetConsoleMode(windll.kernel32.GetStdHandle(-11), 7)

# dimensiona o terminal
print(f'\033[8;24;{TCOLS}t', end='')


def cli_copia_local(folder, data):
    """Info da cópia local."""
    # COPIA LOCAL
    print(COPIA_B + COPIA_F, end='')
    for _linha in range(4):
        print(' ' * TCOLS, end='')
    print('\n\033[4A\033[K', end='')
    print('Pasta da cópia local da PNADC:')
    msg = str(folder)
    if len(str(folder)) > 80:
        msg = str(folder)[:35] + '  (...)  ' + str(folder)[-36:]
    print(COPIA_H + msg)
    msg = aviso = ''
    if data:
        msg = f'Última atualização em {data[1]} (há {elapsed(data[0])})'
        aviso = COPIA_H + 'Cópia local íntegra'
        if data and data[2] is False:
            aviso = COPIA_W + 'CÓPIA LOCAL ADULTERADA!'
    print(COPIA_F + msg)
    print(aviso)


def get_menu(folder):
    """Retorna o menu com os items habilitados."""
    menu = {'1': ('Sincronizar a cópia local com o IBGE', menu_sync),
            '2': ('Organizar e gerar os metadados em formato json',
                  menu_metadados),
            '3': ('Converter os arquivos de microdados',
                  menu_converter),
            '4': ('Montar e identificar os microdados de painéis',
                  menu_paineis),
            'f': ('Definir ou alterar a pasta da cópia local', menu_definir),
            'c': ('Preferências de arquivos de microdados', menu_prefs),
            'h': ('Ajuda', menu_ajuda),
            'q': ('Sair', menu_sair)}
    # selecionar os itens válidos
    if not folder:
        menuitems = ['f', 'h', 'q']
    else:
        with open(Path(folder, LSFILE), encoding='utf-8') as src:
            arquivos = src.read()
        if not arquivos:
            menuitems = ['1', 'f', 'c', 'h', 'q']
        elif not Path(folder, META).is_dir():
            menuitems = ['1', '2', 'f', 'c', 'h', 'q']
        elif not Path(folder, MICRO).is_dir():
            menuitems = ['1', '2', '3', 'f', 'c', 'h', 'q']
        else:
            menuitems = list(menu.keys())
    return {key: value for key, value in menu.items() if key in menuitems}


def get_prefs(folder):
    """Carrega as preferências para microdados."""
    with open(Path(folder, MPFILE), encoding='utf-8') as src:
        return loads(src.read())


def main(folder, data):
    """Inicia pynad."""
    # Apresenta a CLI
    cli_main()
    if not folder and Path(Path.cwd(), IDFILE).is_file():
        folder = Path.cwd()
        data = verify(Path.cwd())
    if not folder:
        folder, data = set_folder(folder)
    # folder pode ter sido ou não definida
    if folder:
        cli_copia_local(folder, data)
    # menu principal
    menu = get_menu(folder)
    print(HEADER_B + HEADER_F)
    print('MENU')
    for key, value in menu.items():
        print(f' {key}) {value[0]}')
    print('\n')
    resp = ''
    while resp not in menu:
        sleep(0.05)
        resp = input(f'{UPCLR}Digite a opção {ARROW} ').lower()
    cli_main()
    return menu[resp][1](folder)


def main_script():
    """Inicia quando roda como script."""
    folder = data = ''
    loop = True
    while loop:
        loop, folder, data = main(folder, data)


def menu_ajuda(folder):
    """Menu item."""
    print('AJUDA\n')
    print('Use a opção "f" para escolher a pasta da cópia local:')
    print('  - a pasta deve conter uma cópia da pnadc criada por pynad')
    print('    ou, para uma nova cópia, deve existir e estar vazia;')
    print('  - não crie arquivos nela ou em suas subpastas,')
    print('    eles serão removidos nas atualizações.\n')
    print('Use a opção "c" para configurar',
          'as opções dos arquivos csv:')
    print('  - se devem ser gerados;\n  - se devem ser compactados.\n')
    print('As opções do menu surgem ao tornarem-se disponíveis:',
          '\n  - execute-as em ordem - 1, 2, 3, 4 -',
          'para sincronizar.\n')
    print('2023 - Rafael Guerreiro Osorio - Ipea',
          '\nEste programa é gratuito e sem garantias,',
          'use por sua própria conta e risco.\n')
    pausar()
    return True, folder, verify(folder)


def menu_converter(folder):
    """Menu item."""
    sync_microdados(folder, get_prefs(folder))
    return True, folder, verify(folder)


def menu_definir(folder):
    """Menu item."""
    newfolder, newdata = set_folder(folder, True)
    if newfolder:
        return True, newfolder, newdata
    return True, folder, verify(folder)


def menu_metadados(folder):
    """Menu item."""
    generate(folder)
    return True, folder, verify(folder)


def menu_paineis(folder):
    """Menu item."""
    sync_paineis(folder, get_prefs(folder))
    return True, folder, verify(folder)


def menu_prefs(folder):
    """Menu item."""
    set_micro_prefs(folder)
    return True, folder, verify(folder)


def menu_sair(folder):
    """Encerra pynad."""
    print('\033c')
    return False, folder, ''


def menu_sync(folder):
    """Menu item."""
    sync(folder)
    return True, folder, verify(folder)


def set_folder(folder, menu=False):
    """Seleciona a pasta da copia local."""

    def seleciona(folder):
        root = tk.Tk()
        root.withdraw()
        return filedialog.askdirectory(
            title='PYNAD - Selecione a pasta da cópia local',
            initialdir=folder,
            mustexist=True)

    # se não há folder, mostrar o menu reduzido
    if not menu:
        return '', ''
    print('DEFINIR OU ALTERAR A PASTA DA CÓPIA LOCAL')
    newfolder = seleciona(folder)
    if newfolder:
        if Path(newfolder, IDFILE).is_file():
            return newfolder, verify(newfolder)
        cli_main()
        cli_copia_local(newfolder, '')
        print(HEADER_B + HEADER_F)
        # verificar se há arquivos na pasta
        # se não houver, é uma nova instalação
        # se tem a pasta COPIA_LOCAL, pode ser uma copia corrompida
        arquivos = [str(arq) for arq in Path(newfolder).glob('**/*')]
        if arquivos:
            print(HEADER_W, end='')
            print('A pasta selecionada contém arquivos!')
            if any(str(Path(newfolder, COPIA_LOCAL))
                   in arq for arq in arquivos):
                print(HEADER_F, end='')
                print('Os arquivos parecem ser de uma cópia local')
                print('Restaurar a cópia_local na pasta selecionada')
                if continuar():
                    register(newfolder)
                    print('Cópia local restaurada. Sincronize-a com o IBGE\n')
                    pausar()
                    return newfolder, verify(newfolder)
            print(HEADER_W, end='')
            print('Não é possível criar uma cópia local nesta pasta')
            print('Por segurança, a pasta deve estar vazia\n')
            print(HEADER_F, end='')
            pausar()
            return folder, verify(folder)
        print('Criar uma cópia local na pasta selecionada')
        if continuar():
            register(newfolder)
            print('Cópia local criada. Sincronize-a com o IBGE\n')
            pausar()
            return newfolder, verify(newfolder)
    return folder, verify(folder)


def set_micro_prefs(folder):
    """
    Define as opções para arquivos csv.

    O formato obrigatório para os microdados é parquet:
      - arquivos csv são gerados para converter os arquivos originais para
        o formato parquet
      - depois de gerados os arquivos parquet, os arquivos csv não são
        mais necessários
      - também não é necessário gerar os arquivos csv com os
        microdados de painel

    Os arquivos csv podem ser compactados.

    As preferências para arquivos csv são guardadas em arquivo
    de configuração e é preciso gerá-lo antes de converter os microdados
    """
    menu = {'A': ('Microdados em formato parquet e csv compactado (zip)',
                  [True, True]),
            'B': ('Microdados em formato parquet e csv descompactado   ',
                  [True, False]),
            'C': ('Somente microdados em formato parquet               ',
                  [False, False])}
    prefs = get_prefs(folder)
    print('DEFINIR AS PREFERÊNCIAS PARA ARQUIVOS DE MICRODADOS\n')
    for key, value in menu.items():
        if value[1] == prefs:
            print(COPIA_B + COPIA_F, end='')
        else:
            print(HEADER_B + HEADER_F, end='')
        print(f' {key}) {value[0]}')
    print(HEADER_B, HEADER_F, '\n')
    resp = ''
    while resp not in menu:
        sleep(0.05)
        resp = input(f'{UPCLR}Digite a nova opção ou ' +
                     f'ENTER para cancelar {ARROW} ').upper()
        # ENTER ou opção atual
        if not resp or (resp in menu and menu[resp][1] == prefs):
            return
    # se há mudança e pasta de microdados já existe
    # remove pasta de microdados (e paineis) após confirmação
    if Path(folder, MICRO).is_dir():
        print(HEADER_W + '\nATENÇÃO: ' + HEADER_F +
              'Alteração de preferências' +
              '\n\nOs arquivos de microdados já convertidos' +
              '\ne os arquivos de painel serão removidos;' +
              '\ntodos terão que ser gerados novamente\n')
        if continuar():
            try:
                rmtree(Path(folder, MICRO))
            except FileNotFoundError:
                pass
            try:
                rmtree(Path(folder, PAINEIS))
            except FileNotFoundError:
                pass
        else:
            return
    with open(Path(folder, MPFILE), 'w', encoding='utf-8') as tgt:
        tgt.write(dumps(menu[resp][1]))


if __name__ == '__main__':
    main_script()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Montagem de microdados de painel da Pnadc.

pynad v3
Ipea - 2023
Rafael Guerreiro Osorio

Para cada visita de um painel podem haver dois arquivos
  1) o arquivo de trimestre básico está contido no de trimestre anual,
     se disponível, usa-se o anual, que tem o suplemento do trimestre
  2) o arquivo com o suplemento da visita
     - o número de observações em uma visita é identico nos arquivos
       de trimestre/trimestral e nos de visita do trimestre
"""

from json import loads, dumps
from pathlib import Path
from time import time
from zipfile import ZipFile, ZIP_DEFLATED
import pyarrow as pa
import pyarrow.compute as pc
import pyarrow.parquet as pq
from pyarrow import csv
from .metadados import META, METALIST
from .converter import MICROPRQ
from .auxiliares import continuar, elapsed, pausar, cli_main
from .paineis_id import identifica_painel

PAINEIS = 'paineis'
PAINEISCSV = Path(PAINEIS, 'csv')
PAINEISPRQ = Path(PAINEIS, 'parquet')
PAINEISMET = Path(PAINEIS, META)
FILESTUB = 'pnadc.microdados.painel.'
METASTUB = 'pnadc.variaveis.painel.'


def available(folder):
    """
    Confere painéis disponíveis e atualizações.

    Tudo deve estar atualizado:
      - sincronização com IBGE
      - metadados
      - arquivos parquet
    """
    # ler o registro de microdados parquet
    with open(Path(folder, MICROPRQ, METALIST), encoding='utf-8') as src:
        regmicro = [loads(reg[:-1]) for reg in src]
    trimestrais = [reg for reg in regmicro
                   if '.trimestral.' in reg['name']]
    a_visitas = [reg for reg in regmicro
                 if '.anual.visita' in reg['name']]
    a_trimestres = [reg for reg in regmicro
                    if '.anual.trimestre' in reg['name']]
    for lista in (trimestrais, a_visitas, a_trimestres):
        lista.sort(key=lambda lista: lista['name'])
    panels = {}
    # Os microdados trimestrais estão ordenados
    # 4 últimos não fazem um painel completo
    for idx, trimestral in enumerate(trimestrais[:-4]):
        ano = int(trimestral['name'].split('.')[-3])
        tri = int(trimestral['name'].split('.')[-2])
        pid = ano * 10 + tri
        panels[pid] = {}
        for vis in range(5):
            panels[pid][f'{vis + 1}'] = [trimestrais[idx + vis]]
            if tri > 4:
                tri = 1
                ano += 1
            # tem suplemento de trimestre? se sim, usar
            sup = [reg for reg in a_trimestres
                   if all(idt in reg['name'] for idt in (
                           f'{ano}', f'trimestre{tri}'))]
            if sup:
                panels[pid][f'{vis + 1}'] = sup[:1]
            # tem suplemento de visita?
            sup = [reg for reg in a_visitas
                   if all(idt in reg['name'] for idt in (
                           f'{ano}', f'visita{vis + 1}'))]
            if sup:
                panels[pid][f'{vis + 1}'].append(sup[0])
            else:
                panels[pid][f'{vis + 1}'].append(None)
            tri += 1
    return panels


def build_cli_main(pid, cur, total, start):
    """Atualiza a CLI."""
    cli_main()
    print('MONTAR E IDENTIFICAR OS MICRODADOS DE PAINEIS\n')
    print(f'Montando paineis - {cur} de {total} - {elapsed(start)}\n')
    print(f'PAINEL {pid}:')


def build_del_tables(panel):
    """Remove tabelas acrescidas por build_panel ao dicionário."""
    for entry in panel.values():
        for item in range(2):
            try:
                del entry[item]['table']
            except (TypeError, IndexError):
                pass


def build_get_tables(folder, pid, panel, start):
    """Lê as tabelas necessárias e lista as variáveis e suas visitas."""
    def trimestre(visita):
        tri = int(f'{pid}'[-1]) - 1 + visita
        if tri > 4:
            tri = tri - 4
        return tri
    order = (('upa', 'ascending'),
             ('v1008', 'ascending'),
             ('v2003', 'ascending'))
    pesos = ('v1028', 'v1032', 'v1036')
    pesosbs = [f'{peso}{rep:0=3}' for peso in pesos for rep in range(1, 201)]
    variaveis = {}
    for key, value in panel.items():
        filtro = pc.field('v1016') == int(key)
        arquivo = Path(folder, MICROPRQ, value[0]['name'])
        build_cli_main(pid, *start)
        print('Obtendo registros da VISITA', key,
              'do arquivo de trimestre\n' +
              value[0]['name'], end='', flush=True)
        # carregar as tabelas do painel como entrada 'table'
        value[0]['table'] = pq.read_table(
            arquivo).filter(filtro).sort_by(order)
        for var in value[0]['table'].column_names:
            if all(var not in lista for lista in (variaveis, pesosbs)):
                variaveis[var] = [{int(key)}, set()]
            elif var not in pesosbs:
                variaveis[var][0].add(int(key))
        if value[1] is not None:
            build_cli_main(pid, *start)
            print('Obtendo registros da VISITA', key,
                  'do arquivo de visitas\n' +
                  value[1]['name'], end='', flush=True)
            filtro = pc.field('trimestre') == trimestre(int(key))
            arquivo = Path(folder, MICROPRQ, value[1]['name'])
            value[1]['table'] = pq.read_table(
                arquivo).filter(filtro).sort_by(order)
            for var in value[1]['table'].column_names:
                if all(var not in lista for lista in (variaveis, pesosbs)):
                    variaveis[var] = [set(), {int(key)}]
                elif var not in pesosbs:
                    variaveis[var][1].add(int(key))
    return build_sort_vars(variaveis)


def build_panel(folder, pid, panel, start):
    """Monta o painel."""
    keys = ['upa', 'v1008', 'v2003', 'v1016']
    comuns, combos, ordem = build_get_tables(folder, pid, panel, start)
    # a tabela do painel começa com UNION (cancat) das tabelas
    # com as variáveis comuns a todas as visitas
    build_cli_main(pid, *start)
    print(f'Montando a tabela do painel\n{len(comuns)}',
          'variáveis comuns', end='', flush=True)
    tabpan = drop_empty_columns(
        pa.concat_tables([table.select(comuns)
                          for table in
                          [panel[f'{vis}'][0]['table']
                           for vis in range(1, 6)]]))
    # para cada combo de variáveis e visitas, union e left join em tabpan
    for combo in ordem:
        build_cli_main(pid, *start)
        print(f'Montando a tabela do painel\n{len(combos[combo])}',
              'variáveis no padrão', combo, end='', flush=True)
        tables = []
        # tabelas de trimestre
        for visita in combo[0]:
            tables.append(panel[f'{visita}'][0]['table'])
        # tabelas de visita
        for visita in combo[1]:
            if visita not in combo[0]:
                tables.append(panel[f'{visita}'][1]['table'])
                # criar v1016 que não vem em arquivo de visita
                tables[-1] = tables[-1].append_column(
                    'v1016', pa.array(
                        (visita,) * len(tables[-1]), pa.int8()))
        for table in tables:
            tabtmp = drop_empty_columns(
                pa.concat_tables([table.select(keys + combos[combo])
                                  for table in tables]))
        tabpan = tabpan.join(tabtmp, keys)
    build_del_tables(panel)
    return tabpan


def build_manager(folder, panels, prefs):
    """Gerencia a conversão."""
    start = time()
    count = 0
    total = len(panels)
    for key, value in panels.items():
        count += 1
        build_cli_main(key, count, total, start)
        table = build_panel(folder, key, value, (count, total, start))
        build_cli_main(key, count, total, start)
        table = identifica_painel(key, table, (count, total, start))
        build_cli_main(key, count, total, start)
        print('Organizando a tabela e gerando metadados', end='', flush=True)
        table = organise_table(table)
        meta = organise_meta(table, value)
        table = to_wide(key, table, (count, total, start))
        if prefs[0] is True:
            build_cli_main(key, count, total, start)
            csvfile = Path(folder, PAINEISCSV, f'{FILESTUB}{key}.csv')
            print('Escrevendo arquivo csv - aguarde...\n' +
                  f'{csvfile.name}', end='', flush=True)
            csv.write_csv(table, csvfile)
            if prefs[1] is True:
                build_cli_main(key, count, total, start)
                compacto = str(csvfile)[:-3] + 'zip'
                print('Compactando arquivo csv - aguarde...\n' +
                      f'{Path(compacto).name}', end='', flush=True)
                with ZipFile(compacto, 'w', ZIP_DEFLATED) as tgt:
                    tgt.write(str(csvfile), csvfile.name)
                csvfile.unlink()
        parquet = Path(folder, PAINEISPRQ, f'{FILESTUB}{key}.parquet')
        build_cli_main(key, count, total, start)
        print('Escrevendo arquivo parquet - aguarde...\n' +
              f'{parquet.name}', end='', flush=True)
        pq.write_table(table, parquet)
        del table
        with open(Path(folder, PAINEISMET, f'{METASTUB}{key}.json'),
                  'w', encoding='utf-8') as tgt:
            tgt.write(dumps(meta))
        regprq = {'name': parquet.name, 'size': parquet.stat().st_size,
                  'files': value}
        with open(Path(folder, PAINEISPRQ, METALIST),
                  'a', encoding='utf-8') as tgt:
            tgt.write(dumps(regprq) + '\n')


def build_sort_vars(variaveis):
    """Separa as variáveis de acordo com os arquivos em que aparecem."""
    combos = {}
    for key, value in variaveis.items():
        entry = (tuple(value[0]), tuple(value[1]))
        if entry not in combos:
            combos[entry] = [key]
        else:
            combos[entry].append(key)
    ordem = list(combos.keys())
    ordem.sort(key=lambda x: len(x[0]), reverse=True)
    comuns = []
    for combo in ordem:
        if combo[0] == (1, 2, 3, 4, 5):
            comuns.extend(combos[combo])
            del combos[combo]
    ordem = list(combos.keys())
    ordem.sort(key=lambda x: len(x[0]), reverse=True)
    return comuns, combos, ordem


def drop_empty_columns(table):
    """Faz o que diz o nome."""
    schema = table.schema
    columns_to_drop = []
    for field in schema:
        column = table[field.name]
        if column.null_count == len(column):
            columns_to_drop.append(field.name)
    return table.drop(columns_to_drop)


def organise_meta(table, panel):
    """Cria metadados."""
    meta = {}
    for key, value in pid_meta().items():
        meta[key] = value
    for visita in panel.values():
        for source in visita:
            if source is not None:
                with open(source['files'][2], encoding='utf-8') as src:
                    variaveis = loads(src.read())
                for var, varmeta in variaveis.items():
                    if var in table.column_names and var not in meta:
                        meta[var] = {'parte': varmeta['parte'],
                                     'quesito': varmeta['quesito'],
                                     'periodo': varmeta['periodo'],
                                     'nome': varmeta['nome'],
                                     'bytes': varmeta['bytes'],
                                     'valores': varmeta['valores']}
    # etapa final para ordenar os metadados
    sorted_meta = {}
    for var in table.column_names:
        sorted_meta[var] = meta[var]
    return sorted_meta


def organise_table(table):
    """Ordena as variáveis e os casos."""
    # ordena moradores por ordem, visita e domicilio
    order = (('upa', 'ascending'),
             ('v1008', 'ascending'),
             ('v1016', 'ascending'),
             ('v2003', 'ascending'))
    # variáveis de identificação do individuo no painel
    painel = ['upa', 'v1008', 'v1016', 'v2001', 'v2003']
    painel.extend([var for var in table.column_names if var[0:3] == 'pid'])
    # outras variáveis de identificação do registro
    comuns = ['ano', 'trimestre', 'uf', 'capital', 'rm_ride',
              'estrato', 'v1014', 'v1022', 'v1023']
    # pesos e projeções de população
    pesos = ['posest', 'posest_sxi']
    pesos.extend([var for var in table.column_names
                  if var[0:3] == 'v10' and int(var[1:]) > 1023
                  and int(var[1:]) < 2000])
    com_v = [var for var in table.column_names if var[0] == 'v'
             and all(var not in lista for lista in (comuns, painel,
                                                    pesos))]
    com_s = [var for var in table.column_names if var[0] == 's']
    resid = [var for var in table.column_names
             if all(var not in lista for lista in (comuns, painel,
                                                   pesos, com_v, com_s))]
    for lista in (pesos, com_v, com_s, resid):
        lista.sort()
    table = table.select(painel + comuns + pesos + com_v +
                         com_s + resid).sort_by(order)
    return table


def pid_meta():
    """Metadados das variáveis pid."""
    from math import log

    def entrevistas(codigo):
        visitas = []
        i = 1
        while i <= codigo:
            # & bitwise and operator
            if i & codigo:
                visitas.append(f'{int(log(i, 2) + 1)}')
            # << left shift
            i <<= 1
        return '-'.join(visitas)
    combos = {}
    for codigo in range(1, 32):
        combos[codigo] = entrevistas(codigo)
    pidcore = (('parte', 'Painel identificado por pynad'), ('quesito', ''),
               ('nome', ''), ('periodo', '1º tri/2012 - atual'), ('bytes', 1),
               ('valores', ''))
    pidvars = (('pid', 'Código do painel - ano e trimestre da 1a visita',
                'AAAAT'),
               ('pidgrp', 'Número sequencial do grupo doméstico', '1 a 5'),
               ('pidind', 'Número sequencial do indivíduo no grupo doméstico',
                '1 ao total de indivíduos'),
               ('pidcla', 'Categorias de identificação', '1 a 7'),
               ('pidgrpent', 'Visitas do grupo doméstico', ''),
               ('pidindent', 'Visitas do indivíduo', ''),
               ('piddnd', 'Dia de nascimento atribuído', '1 a 99'),
               ('piddnm', 'Mês de nascimento atribuído', '1 a 99'),
               ('piddna', 'Ano de nascimento atribuído', 'AAAA'))
    pidmeta = {var[0]: dict(pidcore) for var in pidvars}
    for var in pidvars:
        pidmeta[var[0]]['nome'] = var[1]
        pidmeta[var[0]]['valores'] = var[2]
        if var[0] in ('pid', 'piddna'):
            pidmeta[var[0]]['bytes'] = 4
        if 'ent' in var[0]:
            pidmeta[var[0]]['valores'] = combos
    return pidmeta


def sync_paineis(folder, prefs):
    """Gera e sincroniza painéis."""
    updates, excluir, regprq = verify(folder, prefs)
    if not updates and not excluir:
        print('Painéis atualizados\n')
        pausar()
        return True
    print('Painéis desatualizados!')
    if excluir:
        print(f' - remover {len(excluir)} arquivos de painéis')
    if updates:
        print(f' - montar {len(updates)} painéis')
    if not continuar():
        return False
    start = time()
    for arq in excluir:
        arq.unlink()
    with open(Path(folder, PAINEISPRQ, METALIST),
              'w', encoding='utf-8') as tgt:
        if regprq:
            tgt.write('\n'.join((dumps(reg) for reg in regprq)) + '\n')
    if updates:
        build_manager(folder, updates, prefs)
    cli_main()
    print('MONTAR E IDENTIFICAR OS MICRODADOS DE PAINEIS\n')
    print(f'Painéis atualizados em {elapsed(start)}\n')
    pausar()
    return True


def to_wide(pid, table, start):
    """Reshape da tabela para criar registros de indivíduo."""
    # variáveis constantes nas visitas
    fixas = ['pid', 'uf', 'capital', 'rm_ride', 'v1014', 'v1022', 'v1023',
             'estrato', 'posest', 'upa', 'v1008', 'pidgrp', 'pidgrpent',
             'pidind', 'pidcla', 'pidindent', 'piddnd', 'piddnm', 'piddna']
    # chaves identificadoras do indivíduo
    indkeys = ['upa', 'v1008', 'pidgrp', 'pidind']
    build_cli_main(pid, *start)
    print('Reformatando a tabela para registros de indíviduo')
    print('Variáveis constantes nas visitas', end='', flush=True)
    # a tabela wide começa com as variáveis que não variam por visita
    wide = table.select(fixas).group_by(indkeys).aggregate(
            [(var, 'one') for var in fixas if var not in indkeys])
    wide = wide.rename_columns([var.replace('_one', '') for var
                                in wide.column_names])
    # acrescenta as variáveis de cada visita
    for visita in range(1, 6):
        build_cli_main(pid, *start)
        print('Reformatando a tabela para registros de indíviduo')
        print(f'Variáveis da visita {visita}', end='', flush=True)
        filtro = pc.field('v1016') == visita
        antes = indkeys + [var for var in table.column_names
                           if var not in fixas]
        depois = indkeys + [f'{var}_{visita}' for var in table.column_names
                            if var not in fixas]
        wide = wide.join(drop_empty_columns(
            table.select(antes).filter(filtro).rename_columns(depois)),
            indkeys)
    final = [f'{var}_{visita}' for var in table.column_names
             if var not in fixas for visita in range(1, 6)]
    del table
    final = fixas + [var for var in final if var in wide.column_names]
    build_cli_main(pid, *start)
    print('Reformatando a tabela para registros de indíviduo')
    print('Organizando variáveis e ordenando a tabela', end='', flush=True)
    wide = wide.select(final).sort_by(
        (('upa', 'ascending'), ('v1008', 'ascending'),
         ('pidgrp', 'ascending'), ('pidind', 'ascending')))
    return wide


def verify(folder, prefs):
    """Verifica os arquivos de microdados."""
    # criar as pastas se inexistentes
    if not Path(folder, PAINEISPRQ).is_dir():
        Path(folder, PAINEISPRQ).mkdir(parents=True)
        Path(folder, PAINEISMET).mkdir()
        if prefs[0] is True and not Path(folder, PAINEISCSV).is_dir():
            Path(folder, PAINEISCSV).mkdir()
    verify_meta_csv(folder, prefs)
    # gera a lista de paineis completos nos microdados da Pnadc
    panels = available(folder)
    # carrega as listas de arquivos nas pastas de microdados
    curprq = [{'name': arq.name, 'size': arq.stat().st_size}
              for arq in Path(folder, PAINEISPRQ).glob('*')
              if arq.name != METALIST]
    # carrega os registros dos arquivos
    regprq = []
    if Path(folder, PAINEISPRQ, METALIST).is_file():
        with open(Path(folder, PAINEISPRQ, METALIST), encoding='utf-8') as src:
            regprq = [loads(reg[:-1]) for reg in src]
    # arquivos que existem nas pastas mas não nos registros
    # devem ser excluídos, pois não se sabe os arquivos que lhe originaram
    # arquivos diferentes dos registros também são excluídos
    excluir = list(set(Path(folder, PAINEISPRQ, arq['name'])
                       for arq in curprq if arq not in
                       [{'name': arq['name'], 'size': arq['size']}
                        for arq in regprq]))
    # arquivos que não existem nas pastas ou que serão eliminados
    # precisam ser tirados dos registros
    regprq = [arq for arq in regprq
              if {'name': arq['name'], 'size': arq['size']} in
              curprq]
    # updates se baseiam nos prqs
    updates = {pid: panel for pid, panel in panels.items() if panel not in
               [arq['files'] for arq in regprq]}
    return updates, excluir, regprq


def verify_meta_csv(folder, prefs):
    """Remove arquivo parquet sem metadados e csv."""
    # TODO: deletar arquivos nas pastas csv e metadados
    # que não tenham um arquivo parquet
    pidprq = {arq.name[-13:-8]: arq
              for arq in Path(folder, PAINEISPRQ).glob('*')
              if arq.name != METALIST}
    pidmeta = [arq.name[-10:-5] for arq in Path(folder, PAINEISMET).glob('*')
               if arq.name != METALIST]
    if prefs[0] is True:
        pidcsv = [arq.name[-9:-4]
                  for arq in Path(folder, PAINEISCSV).glob('*')
                  if arq.name != METALIST]
    else:
        pidcsv = list(pidprq.keys())
    for key, value in pidprq.items():
        if key not in pidmeta or key not in pidcsv:
            value.unlink()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Organiza os metadados da Pnadc.

pynad v3
Ipea - 2023
Rafael Guerreiro Osorio

ROTINAS QUE DEPENDEM DE COMO O IBGE DIVULGA METADADOS
"""
import json
import re
import time
import zipfile
from shutil import rmtree
from pathlib import Path
import xlrd
from .copia_local import COPIA_LOCAL, PNADCA, PNADCT, list_local_files
from .auxiliares import elapsed, pausar, CRCLR


DICTRI = 'dicionario_PNADC_microdados_trimestral.xls'


# pastas especificas no IBGE
TRIDOCS = 'Documentacao'
CA_DOCS = 'Documentacao'
CA_DATA = 'Dados'
CA_VIS = 'Visita'
CA_TRI = 'Trimestre'

# nome para pastas das cópias locais e paineis
META = 'metadados'
METALIST = 'pnadc.microdados.dicionarios.json'

# chaves dos dicionários de variáveis json
_VPART = 'parte'  # parte do registro: identificação e controle etc.
_VDESC = 'nome'  # nome da variável
_VPER = 'periodo'  # periodo
_VPOS = 'colunas'  # posição no arquivo COPIA_LOCALginal
_VSIZE = 'bytes'  # bytes necessários para o tipo de dado
_VCAT = 'valores'  # categCOPIA_LOCALas ou valores
_VQUES = 'quesito'  # número da questão para variáveis do questionário
_MISS = 'vazio'  # chave para "não aplicável" em _VCAT


def import_xls(xls):
    """Importa um dicionário de variáveis e retorna em texto TSV."""
    dicionario = xlrd.open_workbook(xls).sheet_by_index(0)
    rows = dicionario.nrows
    cols = dicionario.ncols
    tsv = ''
    for row in range(rows):
        for col in range(cols):
            try:
                datum = str(int(
                    dicionario.cell_value(row, col))).replace('\n', ' ')
            except ValueError:
                datum = str(
                    dicionario.cell_value(row, col)).replace('\n', ' ')
            tsv += datum + '\t'
        tsv = tsv[:-1] + '\n'
    return tsv


def generate(folder):
    """Gera ou atualiza os metadados."""
    print('METADADOS\n')
    start = time.time()
    # é mais prático remover e gerar
    # os metadados do que atualizá-los
    try:
        rmtree(Path(folder, META))
    except FileNotFoundError:
        pass
    finally:
        Path(folder, META).mkdir()
    arquivos = [str(Path(folder, COPIA_LOCAL, arq['path'], arq['name']))
                for arq in list_local_files(folder)]
    anuais = metadata_pnadc_anual(folder, arquivos, start)
    trimestrais = metadata_pnadc_trimestral(folder, arquivos, start)
    microdados = []
    for anual in anuais:
        processados = []
        target = Path(folder, META, Path(anual[1]).name)
        if anual[1] not in processados:
            processados.append(anual[1])
            with open(anual[1], 'rb') as src:
                with open(target, 'wb') as tgt:
                    tgt.write(src.read())
        if anual[2][0] not in processados:
            processados.append(anual[2][0])
            with open(anual[2][0], 'w', encoding='utf-8') as tgt:
                tgt.write(json.dumps(anual[2][1]))
        microdados.append((anual[0], anual[1], anual[2][0]))
    if trimestrais:
        with open(trimestrais[0][2][0], 'w', encoding='utf-8') as tgt:
            tgt.write(json.dumps(trimestrais[0][2][1]))
        for tri in trimestrais:
            microdados.append((tri[0], tri[1], tri[2][0]))
        target = Path(folder, META, METALIST)
        with open(target, 'w', encoding='utf-8') as tgt:
            tgt.write(json.dumps(microdados))
    print(f'{CRCLR}Metadados gerados em {elapsed(start)}\n')
    pausar()


def metadata_pnadc_anual(folder, arquivos, start):
    """
    Extrai e organiza metadados da pnadc anual.

    Esta rotina depende de como o IBGE organiza os
    arquivos para divulgação

    Em dezembro de 2019 a estrutura de disseminação da pnadc
    anual mudou radicalmente

    Há pastas para:
     - 5 visitas (dicionario por visita e por ano)
     - 4 trimestres (dicionário por trimestre vale para todos os anos)
    """
    microdados = metadata_pnadc_anual_arquivos(folder, arquivos)
    for microdado in microdados:
        print(f'{CRCLR}Gerando metadados {elapsed(start)}', end='')
        # lê o dicionário correspondente
        dicionario = metadata_pnadc_vars(import_xls(microdado[1]))
        nome = Path(microdado[0]).name
        if 'trimestre' in nome:
            dicid = nome[:-4].split('_')[2]
        else:
            dicid = '.'.join((nome[:-4].split('_')[2],
                              nome[:-4].split('_')[1]))
        target = f'pnadc.variaveis.anual.{dicid}.json'
        microdado[2] = [str(Path(folder, META, target)), dicionario]
    return microdados


def metadata_pnadc_anual_arquivos(folder, arquivos):
    """Identifica arquivos de microdados anuais."""
    microdados = []
    for vis in range(1, 6):
        datafolder = Path(folder, COPIA_LOCAL, PNADCA, CA_VIS,
                          f'{CA_VIS}_{vis}', CA_DATA)
        docsfolder = Path(folder, COPIA_LOCAL, PNADCA, CA_VIS,
                          f'{CA_VIS}_{vis}', CA_DOCS)
        docfiles = [item for item in arquivos
                    if str(docsfolder) in item
                    and 'dicionario' in item
                    and item.endswith('xls')]
        datafiles = [item for item in arquivos
                     if str(datafolder) in item
                     and item.endswith('zip')]

        # relaciona arquivos de microdados com dicionarios de variaveis
        for doc in docfiles:
            anodoc = Path(doc).name.split('_')[3]
            microdados.extend([[item, doc, '', PNADCA]
                               for item in datafiles
                               if (Path(item).name.split('_')[1]
                                   == anodoc and f'{CA_VIS.lower()}{vis}'
                                   in item) or
                               (Path(item).name.split('_')[1]
                                in ('2013', '2014') and
                                f'{CA_VIS.lower()}{vis}'
                                in item and anodoc == '2012')])
    for tri in range(1, 5):
        datafolder = Path(folder, COPIA_LOCAL, PNADCA, CA_TRI,
                          f'{CA_TRI}_{tri}', CA_DATA)
        docsfolder = Path(folder, COPIA_LOCAL, PNADCA, CA_TRI,
                          f'{CA_TRI}_{tri}', CA_DOCS)
        docfiles = [item for item in arquivos
                    if str(docsfolder) in item
                    and 'dicionario' in item
                    and item.endswith('xls')]
        datafiles = [item for item in arquivos
                     if str(datafolder) in item
                     and item.endswith('zip')]
        if docfiles:
            doc = docfiles[0]
            microdados.extend([[item, doc, '', PNADCA]
                               for item in datafiles])
    return microdados


def metadata_pnadc_trimestral(folder, arquivos, start):
    """
    Extrai e organiza metadados da pnadc trimestral.

    Esta rotina depende de como o IBGE organiza os
    arquivos para divulgação
    """
    # apenas um dicionário para a pnadc trimestral
    archive = [file for file in arquivos if
               all(stub in file for stub
                   in (COPIA_LOCAL, PNADCT, TRIDOCS,
                       'Dicionario_e_input'))][0:1]
    if not archive:
        return []
    with zipfile.ZipFile(archive[0]) as zpf:
        zpf.extract(DICTRI, str(Path(folder, META)))
    dicionario = metadata_pnadc_vars(import_xls(
        str(Path(folder, META, DICTRI))))
    target = str(Path(folder, META, 'pnadc.variaveis.trimestral.json'))

    # lista com os arquivos de microdados trimestrais
    microdados = [[arq, str(Path(folder, META, DICTRI)),
                   [target, dicionario], PNADCT]
                  for arq in arquivos
                  if COPIA_LOCAL in arq and PNADCT in arq and
                  any(parte.isnumeric() for parte in Path(arq).parts)]
    print(f'{CRCLR}Gerando metadados {elapsed(start)}', end='')
    return microdados


def metadata_pnadc_vars(contents):
    """Depende da estrutura dos dicionários de variáveis."""
    def __clean_label(label):
        label = label.replace('"', '').strip()
        label = label.replace('\n', ' ')
        label = re.sub(' +', ' ', label)
        try:
            label = int(float(label))
        except ValueError:
            pass
        return label

    def __set_bytes(size):
        # dtype bytes para armazenar 9 * colunas da var
        # C signed numeric types
        # 99 - 1 byte
        # 9999 - 2 bytes
        # 999999999 - 4 bytes - float ou int
        # >999999999 - 8 bytes - double ou long
        # pnadc pesos são floats - code 15
        bts = 'ERROR'
        if size <= 2:
            bts = 1
        elif size <= 4:
            bts = 2
        elif size <= 9:
            bts = 4
        elif size <= 14:
            bts = 8
        elif size == 15:
            bts = 15
        return bts

    # meta é o dicionário de variáveis
    meta = {}
    curvar = None

    # seção do questionário
    parte = ''

    # pula linhas de cabeçalho e processa
    rows = contents.split('\n')[3:-1]
    for row in rows:

        # line breaks, double spaces e outras coisas
        # limpar campos para processar linhas
        fields = [__clean_label(field) for field in row.split('\t')]

        # linha com informação de "parte" do questionário
        if fields[0] and not fields[1]:
            parte = fields[0].lower()

        # linha principal de variável
        elif all(isinstance(field, int)
                 for field in (fields[0], fields[1])):

            # código (uf, v1008 etc) é a chave em meta
            curvar = fields[2].lower()
            meta[curvar] = {}

            # parte atual
            meta[curvar][_VPART] = parte

            # tuple com:
            # coluna inicial - index em 1
            # coluna final
            # número de colunas
            meta[curvar][_VPOS] = (fields[0],
                                   fields[0] + fields[1] - 1,
                                   fields[1])

            meta[curvar][_VSIZE] = __set_bytes(meta[curvar][_VPOS][2])

            # número do quesito (se tiver)
            meta[curvar][_VQUES] = fields[3]

            # descrição da variável
            meta[curvar][_VDESC] = fields[4].lower()
            if not meta[curvar][_VDESC]:
                meta[curvar][_VDESC] = curvar

            # período
            meta[curvar][_VPER] = fields[7].lower()

            # tem campo 5 - categórica ou info adicional
            meta[curvar][_VCAT] = {}
            especial = (' a ', 'código', 'valor', '130', '01-')
            if (isinstance(fields[5], int)
                    or any(item in fields[5].lower() for item in especial)):
                meta[curvar][_VCAT][fields[5]] = str(fields[6]).lower()
            elif fields[5] or fields[6]:
                meta[curvar][_VCAT] = ', '.join([item.lower()
                                                 for item in fields[5:7]
                                                 if item])

        # linha de categCOPIA_LOCALa
        elif not fields[0] and not fields[1]:
            if not fields[5]:
                fields[5] = _MISS
            try:
                meta[curvar][_VCAT][fields[5]] = fields[6].strip().lower()
            except TypeError:
                # INSPECIONAR AQUI SE ESTIVEREM FALTANDO CATEGCOPIA_LOCALAS
                # print(curvar, fields)
                pass
    return meta


def csv_name(file, keep_ext=False):
    """Retorna o nome do arquivo a partir do original e tipo de pnad."""
    name = str(Path(file).name)
    # originalmente era só csv, mas podem ser zip ou parquet
    ext = name[-name[::-1].find('.'):]
    if str(Path(COPIA_LOCAL, PNADCA)) in file:
        stub = '.'.join((name[:-4].split('_')[2],
                         name[:-4].split('_')[1]))
        pnad = PNADCA
    else:
        stub = name[:-4].split('_')
        stub = f'{stub[1][2:]}.{stub[1][1]}'
        pnad = PNADCT
    if keep_ext:
        return f'pnadc.microdados.{pnad}.{stub}.{ext}'
    return f'pnadc.microdados.{pnad}.{stub}.csv'



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Gerencia cópias locais dos microdados da Pnadc.

pynad v3
Ipea - 2023
Rafael Guerreiro Osorio
"""
from urllib.request import urlopen
from ftplib import FTP, error_perm
from json import dumps, loads
from pathlib import Path, PurePosixPath as pPath
from shutil import disk_usage
from socket import gaierror
from time import strftime, localtime, time
from .auxiliares import continuar, sizestr, elapsed, pausar, cli_main
from .auxiliares import BARRA, CRCLR, UPCLR, HEADER_W, HEADER_F


PNADCA = 'anual'
PNADCT = 'trimestral'
# Servidor FTP IBGE
FTP_FOLDERS = {PNADCA: ('Trabalho_e_Rendimento',
                        'Pesquisa_Nacional_por_Amostra_de_Domicilios_continua',
                        'Anual', 'Microdados'),
               PNADCT: ('Trabalho_e_Rendimento',
                        'Pesquisa_Nacional_por_Amostra_de_Domicilios_continua',
                        'Trimestral', 'Microdados')}
COPIA_LOCAL = 'originais'
IDFILE = str(Path('pynad', 'pynad_data'))
LSFILE = str(Path('pynad', 'pynad_lista'))
LGFILE = str(Path('pynad', 'pynad_sync.log'))
MPFILE = str(Path('pynad', 'pynad_microdados_prefs'))


class AltFTP(FTP):
    """Corrige problema de configuração do servidor de FTP do IBGE."""

    def makepasv(self):
        """
        O servidor do IBGE tem um problema de configuração.

        Retorna um endereço diferente do externo, e em Windows
        isso causa um erro em socket.py, que é chamada por ftplib
        WinError 10060
        """
        _invhost, port = super().makepasv()
        return self.host, port


def download_manager(arquivos, folder):
    """Recebe uma lista de arquivos para download."""
    start = time()
    totalarq = len(arquivos)
    totalbyt = sum(arq['size'] for arq in arquivos)
    print()
    for curarq, arquivo in enumerate(arquivos):
        tgt = Path(folder, COPIA_LOCAL, arquivo['path'], arquivo['name'])
        if not Path(arquivo['path']).parts[1:]:
            src = Path(Path(*FTP_FOLDERS[Path(arquivo['path']).parts[0]]),
                       arquivo['name'])
        else:
            src = Path(Path(*FTP_FOLDERS[Path(arquivo['path']).parts[0]]),
                       '/'.join(Path(arquivo['path']).parts[1:]),
                       arquivo['name'])
        curbyt = sum(arq['size'] for arq in arquivos[:curarq])
        prg = (curarq + 1, totalarq, curbyt, totalbyt, start)
        download_file(src, tgt, prg)
        with open(Path(folder, LSFILE), 'a', encoding='utf-8') as tgt:
            tgt.write(dumps(arquivo) + '\n')
    register(folder)
    print(CRCLR + UPCLR + 'Todos os arquivos foram baixados em',
          elapsed(start), '\n')


def download_file(src, tgt, prg):
    """Baixa um arquivo."""
    def message(src, prg, cum):
        prg_2 = (prg[2] + cum) / prg[3] * 100
        return (f'{CRCLR}{UPCLR}Baixando arquivo {prg[0]} de {prg[1]} - ' +
                f'{prg_2:.0f}% em {elapsed(prg[4])} - ' +
                f'{sizestr((prg[2] + cum)/(time() - prg[4]))}/s' +
                f'\n{src.name}')
    try:
        tgt.parent.mkdir(parents=True)
    except FileExistsError:
        pass

    # baixar com urllib - acrescentar protocolo
    file_url = 'https://ftp.ibge.gov.br/' + str(pPath(src))
    with urlopen(file_url) as response, open(tgt, 'wb') as tar:
        chunk_size = 1024 * 1024
        bytes_downloaded = 0
        print(message(src, prg, bytes_downloaded), end=' ')
        while True:
            chunk = response.read(chunk_size)
            if not chunk:
                break
            tar.write(chunk)
            bytes_downloaded += len(chunk)
            print(message(src, prg, bytes_downloaded), end=' ')


def list_local_files(folder):
    """
    Lista arquivos na copia local.

    Para cada arquivo há um dicionário com 3 entradas:
    - 'parent', 'name' e 'size'
    """
    filelist = [file for file in Path(folder).glob('**/*')
                if file.is_file() and COPIA_LOCAL in file.parts
                and all(part in file.parts for part in Path(folder).parts)]
    arquivos = []
    for file in filelist:
        partes = file.parts
        ini = Path(file).parts.index(COPIA_LOCAL) + 1
        arquivos.append(
            {'path': str(pPath(*partes[ini:-1])),
             'name': partes[-1],
             'size': file.stat().st_size})
    return arquivos


def list_remote_files():
    """
    Lista arquivos de microdados da Pnadc no servidor FTP do IBGE.

    Para cada arquivo há um dicionário com 3 entradas:
        - 'parent', 'name' e 'size'
    """
    def isfolder(path):
        try:
            server.cwd(path)
            server.cwd('..')
            return True
        except error_perm:
            return False

    def mapftp(pnad, ftproot, ftpcur):
        server.cwd(str(pPath(ftproot, ftpcur)))
        itens = server.nlst()
        for item in itens:
            sourcename = pPath(ftproot, ftpcur, item)
            if isfolder(str(sourcename)):
                mapftp(pnad, ftproot, pPath(ftpcur, item))
            else:
                size = server.size(str(sourcename))
                remotefiles.append({'path': str(pPath(pnad, ftpcur)),
                                    'name': item,
                                    'size': size})
                print(f'{UPCLR}Mapeando microdados da Pnadc no servidor' +
                      f' do IBGE - {elapsed(start)}')
                print(f'{CRCLR}{remotefiles[-1]["name"]} ' +
                      f'({sizestr(remotefiles[-1]["size"])})', end='')
    try:
        print('Estabelecendo conexão com o IBGE')
        start = time()
        remotefiles = []
        server = AltFTP('ftp.ibge.gov.br', timeout=60)
        server.connect()
        server.login()
        mapftp(PNADCA, pPath('/', *FTP_FOLDERS[PNADCA]), pPath(''))
        mapftp(PNADCT, pPath('/', *FTP_FOLDERS[PNADCT]), pPath(''))
        server.close()
    except (TimeoutError, gaierror):
        remotefiles = []
    return remotefiles


def register(folder):
    """Registra a cópia local e atualizações criando arquivos."""
    if not Path(folder, 'pynad').is_dir():
        Path(folder, 'pynad').mkdir()
    if not Path(folder, COPIA_LOCAL).is_dir():
        Path(folder, COPIA_LOCAL).mkdir()
    with open(Path(folder, IDFILE), 'w', encoding='utf-8') as tgt:
        tgt.write(str(time()))
    with open(Path(folder, LSFILE), 'w', encoding='utf-8') as tgt:
        for arquivo in list_local_files(folder):
            tgt.write(dumps(arquivo) + '\n')
    # configuração default para microdados
    if not Path(folder, MPFILE).is_file():
        with open(Path(folder, MPFILE), 'w', encoding='utf-8') as tgt:
            tgt.write(dumps((True, True)))


def sync(folder):
    """Sincroniza os arquivos da copia local."""
    print('SINCRONIZAR A CÓPIA LOCAL\n')
    print('Verificar arquivos no servidor do IBGE')
    if not continuar():
        return
    cli_main()
    print('SINCRONIZAR A CÓPIA LOCAL\n')
    arquivos = list_remote_files()
    if not arquivos:
        print(CRCLR + UPCLR + HEADER_W +
              'ERRO DE ACESSO AO SERVIDOR DO IBGE' + HEADER_F)
        print('Causas prováveis:')
        print(' - a conexão com a internet não está ativa')
        print(' - https://ftp.ibge.gov.br está muito lento ou off-line')
        print(' - o firewall não permite conexões nas portas 20000 a 21000\n')
        pausar()
        return
    cli_main()
    print('SINCRONIZAR A CÓPIA LOCAL\n')
    print('Microdados da PNADC no IBGE:', str(len(arquivos)),
          'arquivos:', sizestr(sum(f['size'] for f in arquivos)))
    atuais = list_local_files(folder)
    excluir = [arq for arq in atuais if arq not in arquivos]
    baixar = [arq for arq in arquivos if arq not in atuais]
    if not excluir and not baixar:
        print('A cópia local está sincronizada\n')
        pausar()
        return
    print('A cópia local não está sincronizada!')
    exc_txt = bai_txt = ''
    if excluir:
        exc_txt = (' - remover ' + str(len(excluir)) + ' arquivos: ' +
                   sizestr(sum(f['size'] for f in excluir), 'PT'))
        print(exc_txt)
    if baixar:
        bai_txt = (' - baixar ' + str(len(baixar)) + ' arquivos: ' +
                   sizestr(sum(f['size'] for f in baixar), 'PT'))
        print(bai_txt)
    _total, _used, free = disk_usage(folder)
    print(f'Espaço livre na pasta da cópia local: {sizestr(free)}')
    if not continuar():
        return
    cli_main()
    print('SINCRONIZAR A CÓPIA LOCAL\n')
    for arquivo in excluir:
        Path(folder, COPIA_LOCAL, arquivo['path'], arquivo['name']).unlink()
    register(folder)
    download_manager(baixar, folder)
    write_log(folder, excluir, baixar, exc_txt, bai_txt)
    pausar()


def verify(folder):
    """Verifica a integridade da copia local."""
    if not folder:
        return ''
    integra = False
    with open(Path(folder, IDFILE), encoding='utf-8') as src:
        data = float(src.read())
    datastr = strftime("%Y-%m-%d %H:%M:%S", localtime(data))
    with open(Path(folder, LSFILE), encoding='utf-8') as src:
        lista = src.read()
    arquivos = [loads(item) for item in lista.split('\n')
                if len(item) > 2]
    atuais = list_local_files(folder)
    alterados = [arq for arq in arquivos if arq not in atuais]
    alterados.extend([arq for arq in atuais if arq not in arquivos])
    if not alterados:
        integra = True
    return data, datastr, integra


def write_log(folder, excluir, baixar, exc_txt, bai_txt):
    """Escreve arquivo log com arquivos alterados nas sincronizações."""
    data = verify(folder)
    mode = 'w'
    if Path(folder, LGFILE).is_file():
        mode = 'a'
    with open(Path(folder, LGFILE), mode, encoding='utf-8') as log:
        log.write('Sincronização em ' + data[1] + '\n' + BARRA)
        if excluir:
            log.write('\n' + exc_txt + '\n')
            for arq in excluir:
                log.write(str(Path(arq['path'], arq['name'])) + '\n')
        if baixar:
            log.write('\n' + bai_txt + '\n')
            for arq in baixar:
                log.write(str(Path(arq['path'], arq['name'])) + '\n')
        log.write(BARRA + '\n\n')
